"""
stomp
"""
from __future__ import annotations

import logging
import threading

from proton import Message, Event, Handler, Sender, Receiver
from proton.handlers import MessagingHandler
from proton.reactor import Container

from aspyx.di import on_destroy

from .event import EventManager

class AMQPProvider(MessagingHandler, EventManager.Provider):
    # class property

    logger = logging.getLogger("aspyx.event.amq")  # __name__ = module name

    # local classes

    class AMQHandler(Handler):
        def __init__(self, provider: AMQPProvider):
            super().__init__()

            self.provider = provider

    class AMQPEnvelope(EventManager.Envelope):
        # constructor

        def __init__(self, body="", headers=None):
            self.body = body
            self.headers = headers or {}

        # implement envelope

        def get_body(self) -> str:
            return self.body

        def set(self, key: str, value: str):
            self.headers[key] = value

        def get(self, key: str) -> str:
            return self.headers.get(key,"")

    # constructor

    def __init__(self, server_name: str,  host="localhost", port=61616, user = "", password = ""):
        MessagingHandler.__init__(self)
        EventManager.Provider.__init__(self)

        self.server_name = server_name
        self.host = host
        self.port = port
        self.user = user
        self.password = password

        self.container = Container(self)#, debug=True) # TODO
        self._connection = None

        self.thread= threading.Thread(target=self.container.run, daemon=True)

        self._ready = threading.Event()
        self._senders : dict[str,Sender] = {}  # queue -> sender
        self._receivers : dict[str, Receiver] = {}  # address -> receiver

    # implement MessagingHandler

    def on_transport_error(self, event: Event):
        print(f"[AMQP] Transport error: {event.transport.condition}")

    def on_connection_error(self, event: Event):
        print(f"[AMQP] Connection error: {event.connection.condition}")

    def on_start(self, event: Event):
        self._connection = event.container.connect(
            f"{self.host}:{self.port}",
            user=self.user,
            password=self.password
        )

        self._ready.set()

    def on_connection_closed(self, event: Event):
        self._connection = None

    # internal

    def create_receiver(self, address: str, listener: EventManager.EventListenerDescriptor) -> Receiver:
        class DispatchMessageHandler(MessagingHandler):
            # constructor

            def __init__(self, provider: AMQPProvider, listener: EventManager.EventListenerDescriptor):
                super().__init__()

                self.listener = listener
                self.provider = provider

            # override

            def on_message(self, event: Event):
                self.provider.dispatch(event, self.listener)

        self.logger.info("create receiver %s", address)

        receiver = self.container.create_receiver(self._connection, address, handler=DispatchMessageHandler(self, listener))

        #source = Source(address=address, durable=Terminus.DELIVERIES)
        self._receivers[address] = receiver  # if it exists?

        return receiver

    def dispatch(self, event: Event, listener: EventManager.EventListenerDescriptor):
        AMQPProvider.logger.info("on_message ")

        envelope = self.AMQPEnvelope(body=event.message.body, headers = event.message.properties)

        try:
            self.manager.pipeline.handle(envelope, listener)

            event.delivery.settle()
        except Exception as e:
            pass # already handled


    def get_sender(self, address: str) -> Sender:
        sender = self._senders.get(address, None)
        if not sender:
            self.logger.info("create sender %s", address)

            sender = self.container.create_sender(self._connection, address)

            self._senders[address] = sender

        return sender

    def close_container(self):
        # close all senders

        for sender in self._senders.values():
            try:
                sender.close()
            except Exception as e:
                self.logger.warning("Error closing sender: %s", e)

        # close all receivers

        for receiver in self._receivers.values():
            try:
                receiver.close()
            except Exception as e:
                self.logger.warning("Error closing receiver: %s", e)

        # close connection

        if self._connection:
            try:
                self._connection.close()
            except Exception as e:
                self.logger.warning("Error closing connection: %s", e)

        self.logger.info("AMQPProvider stopped.")

        # stop the container

        self.container.stop()

    def stop(self):
        # local class

        class CloseHandler(AMQPProvider.AMQHandler):
            def __init__(self, provider: AMQPProvider):
                super().__init__(provider)

            def on_timer_task(self, event):
                self.provider.close_container()

        self.container.schedule(0, CloseHandler(self))

    # lifecycle

    @on_destroy()
    def on_destroy(self):
        self.stop()

    # implement Provider

    def start(self):
        self.thread.start()

    def create_envelope(self, body="", headers = None) -> EventManager.Envelope:
        return AMQPProvider.AMQPEnvelope(body=body, headers=headers)

    def listen_to(self, listener: EventManager.EventListenerDescriptor) -> None:
        self._ready.wait(timeout=5)

        class CreateReceiverHandler(AMQPProvider.AMQHandler):
            def __init__(self, provider: AMQPProvider, address: str, listener: EventManager.EventListenerDescriptor):
                super().__init__(provider)

                self.address = address
                self.listener = listener

            def on_timer_task(self, event):
                self.provider.create_receiver(address, self.listener)

        # <event-name>::<listener-name>[-<server-name>]

        address = listener.event.name + "::" + listener.name
        if listener.per_process:
            address = address + f"-{self.server_name}"

        if listener.event.durable:
            address = address + "?durable=true"

        if  self._receivers.get(address, None) is None:
            self.container.schedule(0, CreateReceiverHandler(self, address, listener))

    # implement EnvelopePipeline

    def send(self, envelope: EventManager.Envelope, event_descriptor: EventManager.EventDescriptor):
        # local class

        class SendHandler(AMQPProvider.AMQHandler):
            def __init__(self, provider: AMQPProvider, envelope: AMQPProvider.AMQPEnvelope, address):
                super().__init__(provider)

                self.envelope = envelope
                self.address = address

            def on_timer_task(self, event: Event):
                message = Message(body=self.envelope.get_body(), properties=self.envelope.headers)

                # TODO message.delivery_mode = Message.DeliveryMode.AT_LEAST_ONCE

                self.provider.get_sender(self.address).send(message)

        # go

        address = event_descriptor.name

        self._ready.wait(timeout=5)
        self.container.schedule(0, SendHandler(self, envelope, address))

    def handle(self, envelope: EventManager.Envelope, event_listener_descriptor: EventManager.EventListenerDescriptor):
       self.manager.dispatch_event(event_listener_descriptor, envelope.get_body())



"""
This module provides the core Aspyx event management framework .
"""
from aspyx.di import module
from .event import EventManager, EventListener, event, event_listener, envelope_pipeline, AbstractEnvelopePipeline
from .event_stomp import StompProvider
from .event_amqp import AMQPProvider

@module()
class EventModule:
    def __init__(self):
        pass

__all__ = [
    # event

    "EventModule",

    "EventManager",
    "EventListener",
    "event",
    "event_listener",
    "envelope_pipeline",
    "AbstractEnvelopePipeline",

    # stomp

    "StompProvider",

    # amqp

    "AMQPProvider"
]



"""
test for events
"""
from __future__ import annotations

import logging
import threading

from aspyx.exception import ExceptionManager, handle
from aspyx.util import Logger

from .provider import LocalProvider

Logger.configure(default_level=logging.INFO, levels={
    "httpx": logging.ERROR,
    "aspyx.di": logging.ERROR,
    "aspyx.event": logging.INFO,
    "aspyx.di.aop": logging.ERROR,
    "aspyx.service": logging.ERROR
})

logger = logging.getLogger("test")

logger.setLevel(logging.INFO)

from dataclasses import dataclass

import pytest

from aspyx_event import EventManager, event, envelope_pipeline, AbstractEnvelopePipeline, \
    event_listener, EventListener, EventModule
#StompProvider, AMQPProvider

from aspyx.di import module, Environment, create


# test classes

@dataclass
@event(durable=False)
class HelloEvent:
    hello: str

@envelope_pipeline()
class SessionPipeline(AbstractEnvelopePipeline):
    # constructor

    def __init__(self):
        super().__init__()

    # implement

    def send(self, envelope: EventManager.Envelope, event_descriptor: EventManager.EventDescriptor):
        envelope.set("session", "session")

        self.proceed_send(envelope, event_descriptor)

    def handle(self, envelope: EventManager.Envelope, event_listener_descriptor: EventManager.EventListenerDescriptor):
        session = envelope.get("session")

        self.proceed_handle(envelope, event_listener_descriptor)


sync_event_received = threading.Event()
async_event_received = threading.Event()

@event_listener(HelloEvent, per_process=True)
class SyncListener(EventListener[HelloEvent]):
    received = None

    # constructor

    def __init__(self):
        pass

    # implement

    def on(self, event: HelloEvent):
        SyncListener.received = event

        sync_event_received.set()

@event_listener(HelloEvent, per_process=True)
class AsyncListener(EventListener[HelloEvent]):
    received = None

    # constructor

    def __init__(self):
        pass

    # implement

    def on(self, event: HelloEvent):
        AsyncListener.received = event

        async_event_received.set()

# test module

@module(imports=[EventModule])
class Module:
    # constructor

    def __init__(self):
        pass

    # handlers

    @handle()
    def handle_exception(self, exception: Exception):
        print(exception)

    # internal

    def create_exception_manager(self):
        exception_manager = ExceptionManager()

        exception_manager.collect_handlers(self)

        return exception_manager

    @create()
    def create_event_manager(self) -> EventManager:
        return EventManager(LocalProvider(), exception_manager=self.create_exception_manager())
        # EventManager(StompProvider(host="localhost", port=61616, user="artemis", password="artemis"))
        # EventManager(AMQPProvider("server-id", host="localhost", port=5672, user="artemis", password="artemis"))

@pytest.fixture(scope="session")
def environment():
    environment = Environment(Module)  # start server

    yield environment

    environment.destroy()

class TestLocalService():
    def test_events(self, environment):
        event_manager = environment.get(EventManager)

        event = HelloEvent("world")

        event_manager.send_event(event)

        assert sync_event_received.wait(timeout=1), "sync event not received"
        assert async_event_received.wait(timeout=1), "async event not received"

        assert event == SyncListener.received, "events not =="
        assert event == AsyncListener.received, "events not =="



import logging
from aspyx.util import Logger

Logger.configure(default_level=logging.NOTSET, levels={
    "aspyx.di": logging.ERROR,
    "aspyx.event": logging.INFO,
    "aspyx.di.aop": logging.ERROR,
    "aspyx.service": logging.ERROR
})

from aspyx_event import EventManager, AMQPProvider

#from packages.aspyx_event.performance.common import Module, HelloEvent

###

"""
test for health checks
"""

import asyncio
import logging

from aspyx.util import Logger


Logger.configure(default_level=logging.INFO, levels={
    "httpx": logging.ERROR,
    "aspyx.di": logging.ERROR,
    "aspyx.event": logging.INFO,
    "aspyx.di.aop": logging.ERROR,
    "aspyx.service": logging.ERROR
})

logger = logging.getLogger("test")

logger.setLevel(logging.INFO)

from dataclasses import dataclass

from aspyx_event import EventManager, event, envelope_pipeline, AbstractEnvelopePipeline, \
    event_listener, EventListener, EventModule
#StompProvider, AMQPProvider

from aspyx.di import module, Environment, create


# test classes

@dataclass
@event(durable=False)
class HelloEvent:
    hello: str

@envelope_pipeline()
class SessionPipeline(AbstractEnvelopePipeline):
    # constructor

    def __init__(self):
        super().__init__()

    # implement

    def send(self, envelope: EventManager.Envelope, event_descriptor: EventManager.EventDescriptor):
        envelope.set("session", "session")

        self.proceed_send(envelope, event_descriptor)

    def handle(self, envelope: EventManager.Envelope, event_listener_descriptor: EventManager.EventListenerDescriptor):
        session = envelope.get("session")

        self.proceed_handle(envelope, event_listener_descriptor)


@event_listener(HelloEvent, per_process=True)
class SyncListener(EventListener[HelloEvent]):
    received = None

    # constructor

    def __init__(self):
        pass

    # implement

    def on(self, event: HelloEvent):
        print(".")

@event_listener(HelloEvent, per_process=True)
class AsyncListener(EventListener[HelloEvent]):
    received = None

    # constructor

    def __init__(self):
        pass

    # implement

    def on(self, event: HelloEvent):
        print(".")

# test module

@module(imports=[EventModule])
class Module:
    def __init__(self):
        pass

    @create()
    def create_event_manager(self) -> EventManager:
        #return EventManager(LocalProvider())
        #return EventManager(StompProvider(host="localhost", port=61616, user="artemis", password="artemis"))
        return EventManager(AMQPProvider("server-id", host="localhost", port=5672, user="artemis", password="artemis"))

###

#from .common import Module, HelloEvent




async def main():
    environment = Environment(Module)
    event_manager = environment.get(EventManager)

    event = HelloEvent("world")

    event_manager.send_event(event)

    while True:
        event_manager.send_event(event)
        await asyncio.sleep(0.1)

if __name__ == "__main__":
    asyncio.run(main())


"""
event management
"""
from __future__ import annotations

import asyncio
import json
import logging

from abc import ABC, abstractmethod
from typing import Type, TypeVar, Generic, Any, Optional

from aspyx.exception import ExceptionManager
from aspyx.reflection import Decorators

from aspyx.di import Environment, inject_environment, Providers, ClassInstanceProvider, on_destroy

from aspyx_service.serialization import get_deserializer, get_serializer

class EventException(Exception):
    pass

T = TypeVar("T")

class EventListener(Generic[T]):
    """
    An `EventListener` listens to a single event.
    """
    def on(self, event: T) -> None:
        """
        Callback when an event occurs.

        Args:
            event: the event
        """
        pass

class EventManager:
    """
    Central class that manages sending and receiving/dispatching events.
    """
    # local classes

    class EventDescriptor:
        """
        Covers the meta-data of an event.
        """
        def __init__(self, type: Type):
            self.type = type

            args = Decorators.get_decorator(type, event).args

            self.name = args[0]
            if self.name == "":
                self.name = type.__name__

            self.broadcast : bool = args[1]
            self.durable : bool   = args[2]

    class EventListenerDescriptor:
        """
       Covers the meta-data of an event listener.
       """
        def __init__(self, type: Type, event_type: Type, name: str, group: str, per_process: bool):
            if name == "":
                name = type.__name__

            self.type : Type = type
            self.name = name
            self.event = EventManager.EventDescriptor(event_type)

            self.group = group
            self.per_process = per_process

    class Envelope(ABC):
        """
        Wrapper around an event while being received or sent.
        """
        @abstractmethod
        def get_body(self) -> str:
            """
            return the body as a str

            Returns:
                str: the body
            """
            pass

        @abstractmethod
        def set(self, key: str, value: str) -> None:
            """
            set a header value

            Args:
                key: a key
                value: the value
            """

        @abstractmethod
        def get(self, key: str) -> str:
            """
            retrieve a header value

            Args:
                key: a key

            Returns:
                str: the value
           """

    class EnvelopePipeline(ABC):
        """
        An interceptor for sending and receiving events
        """
        @abstractmethod
        def send(self, envelope: EventManager.Envelope, event_descriptor: EventManager.EventDescriptor):
            """
            interceptor on the sending side
            Args:
                envelope: the envelope
                event_descriptor: the event descriptor
            """

        @abstractmethod
        def handle(self, envelope: EventManager.Envelope, event_listener_descriptor: EventManager.EventListenerDescriptor):
            """
            interceptor on the handling side
            Args:
                envelope: the envelope
                event_listener_descriptor: the listener descriptor
            """

    class Provider(EnvelopePipeline):
        """
        The bridge to a low-level queuing library.
        """
        # constructor

        def __init__(self):
            self.manager : Optional[EventManager] = None

        # abstract

        def start(self):
            pass

        def stop(self):
            pass

        @abstractmethod
        def create_envelope(self, body="", headers = None) -> EventManager.Envelope:
            pass

        @abstractmethod
        def listen_to(self, listener: EventManager. EventListenerDescriptor) -> None:
            pass

    # class properties

    logger = logging.getLogger("aspyx.event")  # __name__ = module name

    pipelines: list[Type] = []

    events: dict[Type, EventDescriptor] = {}
    event_listeners: list[EventManager.EventListenerDescriptor] = []

    events_by_name: dict[str, EventDescriptor] = {}

    # class methods

    @classmethod
    def register_envelope_pipeline(cls, handler: Type):
        cls.pipelines.append(handler)

    @classmethod
    def register_event(cls, descriptor: EventManager.EventDescriptor):
        cls.events[descriptor.type] = descriptor

        cls.events_by_name[descriptor.name] = descriptor

    @classmethod
    def register_event_listener(cls, descriptor: EventManager.EventListenerDescriptor):
        cls.event_listeners.append(descriptor)

    # constructor

    def __init__(self, provider: EventManager.Provider, exception_manager: Optional[ExceptionManager] = None):
        """
        create a new `EventManager`

        Args:
            provider: an `EventManager.Provider`
        """
        self.environment : Optional[Environment] = None
        self.provider = provider
        self.pipeline = self.provider
        self.exception_manager = exception_manager

        provider.manager = self

        self.loop = asyncio.get_event_loop()

        self.setup()

    # inject

    @inject_environment()
    def set_environment(self, environment: Environment):
        self.environment = environment

        # create & chain pipelines

        for type in self.pipelines:
            pipeline = environment.get(type)

            if isinstance(pipeline, AbstractEnvelopePipeline):
                pipeline.next = self.pipeline

            self.pipeline = pipeline

    # lifecycle

    @on_destroy()
    def on_destroy(self):
        self.provider.stop()

    # internal

    def get_event_descriptor(self, type: Type) -> EventManager.EventDescriptor:
        descriptor =  self.events.get(type, None)

        if descriptor is None:
            raise EventException(f"{type.__name__} is not an event")

        return descriptor

    def listen_to(self, listener: EventManager.EventListenerDescriptor):
        self.provider.listen_to(listener)

    def setup(self):
        # start

        self.provider.start()

        # listeners

        for listener in self.event_listeners:
            # replace initial object

            listener.event = self.get_event_descriptor(listener.event.type)

            # install listener

            self.listen_to(listener)


    def get_listener(self, type: Type) -> Optional[EventListener]:
        return self.environment.get(type)

    def to_json(self, obj) -> str:
        dict = get_serializer(type(obj))(obj)

        return json.dumps(dict)

    def dispatch_event(self, descriptor: EventManager.EventListenerDescriptor, body: Any):
        event = get_deserializer(descriptor.event.type)(json.loads(body))

        listener = self.get_listener(descriptor.type)

        try:
            listener.on(event)
        except Exception as e:
            if self.exception_manager is not None:
                raise self.exception_manager.handle(e)

            self.logger.error(f"caught an exception: {e} while dispatching event {descriptor.event.name}")
            raise e

        #async def call_handler(listener, event):
        #    if inspect.iscoroutinefunction(listener.on):
        #        await listener.on(event)
        #    else:
        #        listener.on(event)

        # schedule handler in main loop

        #self.loop.call_soon(asyncio.create_task, call_handler(listener, event))
        #asyncio.create_task(call_handler(listener, event))

    # public

    def send_event(self, event: Any) -> None:
        """
        send an event.

        Args:
            event: the event
        """
        descriptor = self.get_event_descriptor(type(event))

        envelope = self.provider.create_envelope(body=self.to_json(event), headers={})

        self.pipeline.send(envelope, descriptor)

def event(name="", broadcast=False, durable=False):
    """
    decorates event classes

    Args:
        name: the event name
        durable: if `True`, the corresponding queue is persistent

    Returns:

    """
    def decorator(cls):
        Decorators.add(cls, event, name, broadcast, durable)

        EventManager.register_event(EventManager.EventDescriptor(cls))

        return cls

    return decorator

def event_listener(event: Type, name="", group="", per_process = False):
    """
    decorates event listeners.

    Args:
        event: the event type
        name: the listener name
        per_process: if `True`, listeners will process events on different processes

    Returns:

    """
    def decorator(cls):
        Decorators.add(cls, event_listener, event, name, group, per_process)

        EventManager.register_event_listener(EventManager.EventListenerDescriptor(cls, event, name, group, per_process))
        Providers.register(ClassInstanceProvider(cls, False))

        return cls

    return decorator

def envelope_pipeline():
    """
    decorates an envelope pipeline
    """
    def decorator(cls):
        Decorators.add(cls, envelope_pipeline)

        EventManager.register_envelope_pipeline(cls)
        Providers.register(ClassInstanceProvider(cls, True, "singleton"))

        return cls

    return decorator

class AbstractEnvelopePipeline(EventManager.EnvelopePipeline):
    """
    abstract base-class for envelope pipelines
    """
    # constructor

    def __init__(self, envelope_handler: Optional[EventManager.EnvelopePipeline] = None):
        self.next = envelope_handler

    # public

    def proceed_send(self, envelope: EventManager.Envelope, event_descriptor: EventManager.EventDescriptor):
        self.next.send(envelope, event_descriptor)

    def proceed_handle(self, envelope: EventManager.Envelope, descriptor: EventManager.EventListenerDescriptor):
        self.next.handle(envelope, descriptor)


"""
stomp provider
"""
from __future__ import annotations

from typing import Optional
import stomp

from aspyx.di import on_destroy

from .event import EventManager


class StompProvider(EventManager.Provider, stomp.ConnectionListener):
    # local classes

    class StompEnvelope(EventManager.Envelope):
        # constructor

        def __init__(self, body="", headers=None):
            self.body = body
            self.headers = headers or {}

        # implement envelope

        def get_body(self) -> str:
            return self.body

        def set(self, key: str, value: str):
            self.headers[key] = value

        def get(self, key: str) -> str:
            return self.headers.get(key,"")

    # constructor

    def __init__(self, host="localhost", port=61616, user = "", password = ""):
        super().__init__()

        self.host = host
        self.port = port
        self.user = user
        self.password = password

        self.connection : Optional[stomp.Connection] = self.connect()

    # implement ConnectionListener

    def on_connected(self, frame):
        print('Connected to broker') # TODO

    def get_headers(self, frame):
        headers = frame.headers
        if isinstance(headers, dict):
            return headers
        else:
            return dict(headers)

    def on_message(self, frame):
        print(frame)

        headers = self.get_headers(frame)

        message_id = headers.get("message-id")
        destination = headers.get("destination")
        body = frame.body

        event_name = destination.split("/")[-1]

        event_descriptor = EventManager.events_by_name.get(event_name, None)

        envelope = self.create_envelope(body=body, headers=headers)

        self.manager.pipeline.handle(envelope, event_descriptor)

    # lifecycle

    @on_destroy()
    def on_destroy(self):
        if self.connection:
            self.connection.disconnect()
            self.connection = None

    # internal

    def connect(self) -> stomp.Connection:
        connection = stomp.Connection([(self.host, self.port)])

        connection.set_listener('', self)

        connection.connect(self.user, self.password, wait=True)

        return connection

    # implement Provider

    def create_envelope(self, body="", headers = None) -> EventManager.Envelope:
        return StompProvider.StompEnvelope(body=body, headers=headers)

    def listen_to(self, listener: EventManager.EventListenerDescriptor) -> None:
        self.connection.subscribe(destination=f"/queue/{listener.event.name}", id=f"id-{listener.event.name}", ack="auto")

    # implement EnvelopePipeline

    def send(self, envelope: EventManager.Envelope, event_descriptor: EventManager.EventDescriptor):
        self.connection.send(body=envelope.get_body(), destination=f"/queue/{event_descriptor.name}")

    def handle(self, envelope: EventManager.Envelope, event_listener_descriptor: EventManager.EventListenerDescriptor):
        self.manager.dispatch_event(event_listener_descriptor, envelope.get_body())



from typing import Optional

from aspyx.di import Environment, inject_environment
from aspyx_event import EventManager


class LocalProvider(EventManager.Provider):
    # local classes

    class TestEnvelope(EventManager.Envelope):
        # constructor

        def __init__(self, body="", headers=None):
            self.body = body
            self.headers = headers or {}

        # implement envelope

        def get_body(self) -> str:
            return self.body

        def set(self, key: str, value: str):
            self.headers[key] = value

        def get(self, key: str) -> str:
            return self.headers.get(key,"")

    # constructor

    def __init__(self):
        super().__init__()

        self.environment : Optional[Environment] = None
        self.listeners : list[EventManager.EventListenerDescriptor] = []

    # inject

    @inject_environment()
    def set_environment(self, environment: Environment):
        self.environment = environment

    # implement Provider

    def create_envelope(self, body="", headers = None) -> EventManager.Envelope:
        return LocalProvider.TestEnvelope(body=body, headers=headers)

    def listen_to(self, listener: EventManager.EventListenerDescriptor) -> None:
        self.listeners.append(listener)

    # implement EnvelopePipeline

    def send(self, envelope: EventManager.Envelope, event_descriptor: EventManager.EventDescriptor):
        #self.handle(envelope, event_descriptor)
        self.manager.pipeline.handle(envelope, event_descriptor)

    def handle(self, envelope: EventManager.Envelope, event_descriptor: EventManager.EventDescriptor):
        for listener in self.listeners:
            if listener.event is event_descriptor:
                self.manager.dispatch_event(listener, envelope.get_body())
